;this is the lzsa decompression tool by Emmanuel Marty
;https//github.com/emmanuel-marty/lzsa
;files have to be compressed with lzsa.exe -f2 -r <sourcefile> <outfile>

DecompressLZSA2:
	; in many places we assume that B = 0
	; flag P in A' signals the need to re-load the nibble store
	xor a
	ld b,a
	ex af,af'
	jr ReadToken

CASE00x:
	; token "00Z" stands for 5-bit offsets
	; (read a nibble for offset bits 1-4 and use the inverted bit Z
	; of the token as bit 0 of the offset; set bits 5-15 of the offset to 1)
	push af
	call skipLDCA
	ld c,a
	pop af
	cp %00100000
	rl c
	jr SaveOffset

CASE0xx:
	dec b
	cp %01000000
	jr c,CASE00x

CASE01x:
	; token "01Z" stands for 9-bit offsets
	; (read a byte for offset bits 0-7 and use the inverted bit Z
	; for bit 8 of the offset; set bits 9-15 of the offset to 1)
	cp %01100000
doRLB:
	rl b

OffsetReadC:
	ld c,(hl)
	inc hl

SaveOffset:
	ld (PrevOffset),bc
	ld b,0

MatchLen:
	and %00000111
	add 2
	cp 9
	call z,ExtendedCode

CopyMatch:
	ld c,a
	push hl		; BC = len, DE = dest, HL = -offset, SP -> [src]

PrevOffset EQU $+1
	ld hl,0
	add hl,de
	ldir					; BC = 0, DE = dest
	pop hl						; HL = src

ReadToken:
	ld a,(hl)
	inc hl
	push af
	and %00011000
	jr z,NoLiterals
	rrca
	rrca
	rrca
	call pe,ExtendedCode
	ld c,a
	ldir
NoLiterals:
	pop af
	or a
	jp p,CASE0xx

CASE1xx:
	cp %11000000
	jr c,CASE10x
		; token "111" stands for repeat offsets
		; (reuse the offset value of the previous match command)
	cp %11100000
	jr nc,MatchLen

CASE110:		; token "110" stands for 16-bit offset
	; (read a byte for offset bits 8-15, then another byte for offset bits 0-7)
	ld b,(hl)
	inc hl
	jr OffsetReadC

CASE10x:		; token "10Z" stands for 13-bit offsets
	; (read a nibble for offset bits 9-12 and use the inverted bit Z
	; for bit 8 of the offset, then read a byte for offset bits 0-7.
	; set bits 13-15 of the offset to 1. substract 512 from the offset to get the final value)
	call ReadNibble
	ld b,a
	ld a,c
	cp %10100000
	dec b
	jr doRLB

ExtendedCode:
	call ReadNibble
	inc a
	jr z,ExtraByte
	sub #F0+1
	add c
	ret
ExtraByte:
	ld a,15
	add c
	add (hl)
	inc hl
	ret nc
	ld a,(hl)
	inc hl
	ld b,(hl)
	inc hl
ret nz
	pop bc		; RET is not needed, because RET from ReadNibble is sufficient

ReadNibble:
	ld c,a
skipLDCA:
	xor a
	ex af,af'
	ret m
	ld a,(hl)
	or #F0
	ex af,af'
	ld a,(hl)
	inc hl
	or #0F
	rrca
	rrca
	rrca
	rrca
	ret
